#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


import code
import logging
import os
import signal
import sys

from xoeuf import odoo
from xoeuf.odoo.tools import config

from xoeuf.api import contextual
from xoutil.future.types import new_class


from . import Command

_logger = logging.getLogger(__name__)


def raise_keyboard_interrupt(*a):
    raise KeyboardInterrupt()


class Console(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<console>"):
        code.InteractiveConsole.__init__(self, locals, filename)
        try:
            import readline
            import rlcompleter
        except ImportError:
            print('readline or rlcompleter not available, '
                  'autocomplete disabled.')
        else:
            readline.set_completer(rlcompleter.Completer(locals).complete)
            readline.parse_and_bind("tab: complete")


class Base(object):
    # This Base is just here to have many 'shell' commands.  See aliases
    # below.  Odoo 9 includes its own 'shell' command, we the aliases to be
    # able to work with our shell.
    #
    # In fact, this is a backport of Odoo 10 shell command into xoeuf, so that
    # our shell for Odoo 8, behaves the same as theirs.
    supported_shells = ['ipython', 'ptpython', 'bpython', 'python']

    def init(self, args):
        try:
            pos = args.index('--')
            args, self.shell_args = args[:pos], args[pos+1:]  # noqa: E226
        except ValueError:
            self.shell_args = []
        config.parse_config(args)
        odoo.cli.server.report_configuration()
        odoo.service.server.start(preload=[], stop=True)
        signal.signal(signal.SIGINT, raise_keyboard_interrupt)

    def console(self, local_vars):
        from xoutil.eight import exec_
        if not os.isatty(sys.stdin.fileno()):
            exec_(sys.stdin, local_vars)
        else:
            if 'env' not in local_vars:
                print(
                    'No environment set, use `%s shell -d dbname` '
                    'to get one.' % sys.argv[0]
                )
            for i in sorted(local_vars):
                print('%s: %s' % (i, local_vars[i]))

            preferred_interface = config.options.get('shell_interface')
            if preferred_interface:
                shells_to_try = [preferred_interface, 'python']
            else:
                shells_to_try = self.supported_shells

            for shell in shells_to_try:
                try:
                    return getattr(self, shell)(local_vars)
                except ImportError:
                    pass
                except Exception:
                    _logger.warning("Could not start '%s' shell." % shell)
                    _logger.debug("Shell error:", exc_info=True)

    def ipython(self, local_vars):
        from IPython import start_ipython
        start_ipython(argv=self.shell_args, user_ns=local_vars)

    def ptpython(self, local_vars):
        from ptpython.repl import embed
        embed({}, local_vars)

    def bpython(self, local_vars):
        from bpython import embed
        embed(local_vars)

    def python(self, local_vars):
        Console(locals=local_vars).interact()

    def shell(self, dbname):
        local_vars = {
            'openerp': odoo,
            'odoo': odoo,
        }
        if dbname:
            registry = odoo.registry(dbname)
            with registry.cursor() as cr:
                uid = odoo.SUPERUSER_ID
                ctx = odoo.api.Environment(cr, uid, {})['res.users'].context_get()  # noqa
                env = odoo.api.Environment(cr, uid, ctx)
                local_vars['env'] = env
                local_vars['self'] = env.user
                self.console(local_vars)
        else:
            self.console(local_vars)

    @contextual
    def run(self, args):
        self.init(args)
        self.shell(config['db_name'])
        return 0


for alias in ('shell', 'ishell', 'python', 'ipython'):
    class aliased(object):
        command_cli_name = alias

    def exec_body(d):
        d['__doc__'] = 'A shell for Odoo'
        return d

    globals()[alias] = new_class(
        alias,
        (aliased, Base, Command),
        exec_body=exec_body
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf ORM extensions for Open Object (OpenERP) models.

All the attributes of this module must be functions that can be integrated as
`BaseModel` methods or operators.

Operators name has the format ``operator__<name>`` and are converted to
``__<name>__``.

Use :func:`xoeuf.osv.improve.integrate_extensions` for integrate these
extensions.

'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import)

from xoeuf.models import _past
_past.dissuade()

from xoutil.deprecation import deprecated

del _past


INTEGRATED = False    # True after integrated to `ModelBase`


def search_read(self, cr, uid, *args, **kwargs):
    '''Search based on a domain and with the returned ids read records with the
    given fields.

    :param self: model to operate in
    :param cr: database cursor
    :param uid: current user id

    Other optional arguments can be passed by position or by name:

    - ``domain``: list of tuples specifying the search domain (see below).  An
      empty list or no argument can be used to match all records.  Could be
      passed by position after ``uid``. Use ``args`` as an alias in arguments
      by name (``kwargs``).

    - ``fields``: list of field names to return (default: all fields would be
      returned). Example ``['id', 'name', 'age']``.  Could be passed by
      position after ``domain``.

    - ``context``: context arguments in a dictionary, like lang, time
      zone. Could be passed by position after ``fields``.

    Keyword only arguments:

    - ``offset``: number of results to skip in the returned values
      (default: ``0``).

    - ``limit``: max number of records to return (default: unlimited)

    - ``order``: columns to sort by (default: ``self._order=id``)

    - ``ensure_list``:  Always return a list.  By default is False.

      .. versionadded:: 0.5.1

    :return: dictionary or list of dictionaries (one per record asked) with
             requested field values.
    :rtype: ``[{‘name_of_the_field’: value, ...}, ...]``

    :raise AccessError:
      * if user has no read rights on the requested object, or
      * if user tries to bypass access rules for read on the requested object

    **Expressing a search domain (or 'args')**

    Each tuple in the search domain needs to have 3 elements, in the form:
      ``('field_name', 'operator', value)``, where:

        - **field_name** must be a valid name of field of the object model,
            possibly following many-to-one relationships using dot-notation,
            e.g 'street' or 'partner_id.country' are valid values.

        - **operator** must be a string with a valid comparison operator from
            this list: ``=``, ``!=``, ``>``, ``>=``, ``<``, ``<=``, ``like``,
            ``ilike``, ``in``, ``not in``, ``child_of``, ``parent_left``,
            ``parent_right``.

          The semantics of most of these operators are obvious.  The
          ``child_of`` operator will look for records who are children or
          grand-children of a given record, according to the semantics of this
          model (i.e following the relationship field named by
          ``self._parent_name``, by default ``parent_id``.

        - **value** must be a valid value to compare with the values of
          ``field_name``, depending on its type.

        Domain criteria can be combined using 3 logical operators than can be
        added between tuples: ``&`` (logical AND, default), ``|`` (logical
        OR), ``!`` (logical NOT).  These are *prefix* operators and the arity
        of the ``&`` and ``|`` operator is 2, while the arity of the ``!`` is
        just 1.  Be very careful about this when you combine them the first
        time.

        Here is an example of searching for Partners named *ABC* from Belgium
        or Germany whose language is not English::

            [('name','=','ABC'),'!',('language.code','=','en_US'),
             '|',('country_id.code','=','be'),('country_id.code','=','de'))

        The ``&`` is omitted as it is the default, and of course we could have
        used ``!=`` for the language, but what this domain really represents
        is::

            (name is 'ABC' AND (language is NOT english) AND
             (country is Belgium OR Germany))

    '''
    from xoutil.objects import pop_first_of as _get
    # Convert all positional to keyword arguments
    for pos, arg in enumerate(args):
        kwargs[pos + 3] = arg
    # Get all arguments or default values
    domain = _get(kwargs, 3, 'domain', 'args', default=[])
    fields = _get(kwargs, 4, 'fields', default=[])
    ctx = _get(kwargs, 5, 'context', default={})
    offset = _get(kwargs, 'offset', default=0)
    limit = _get(kwargs, 'limit', default=None)
    order = _get(kwargs, 'order', default=None)
    ensure_list = _get(kwargs, 'ensure_list', default=False)
    assert not kwargs, \
        "Invalid %s arguments: %s" % (len(kwargs), kwargs.keys())
    # Do it
    ids = self.search(cr, uid, domain, offset=offset, limit=limit,
                      order=order, context=ctx)
    if len(ids) == 1 and not ensure_list:
        ids = ids[0]
    return self.read(cr, uid, ids, fields=fields, context=ctx) if ids else []


@deprecated('search', 'Use the new API search method')
def search_browse(self, cr, uid, *args, **kwargs):
    '''Search based on a domain and with the returned ids browse corresponding
    records or return None if nothing is found.

    :param self: model to operate in
    :param cr: database cursor
    :param uid: current user id

    The rest of the arguments are the same as `search_read`:func:, excluding
    `fields` which is not acceptable in this method.

    :return: object or list of objects requested or None

    :raise AccessError:
      * if user tries to bypass access rules for read on the requested object

    See :func:`search_read` for how to express a search domain.

    '''
    from xoutil.objects import pop_first_of as _get
    # Convert all positional to keyword arguments
    for pos, arg in enumerate(args):
        kwargs[pos + 3] = arg
    # Get all arguments or default values
    domain = _get(kwargs, 3, 'domain', 'args', default=[])
    ctx = _get(kwargs, 4, 'context', default={})
    offset = _get(kwargs, 'offset', default=0)
    limit = _get(kwargs, 'limit', default=None)
    order = _get(kwargs, 'order', default=None)
    ensure_list = _get(kwargs, 'ensure_list', default=False)
    assert not kwargs, \
        "Invalid %s arguments: %s" % (len(kwargs), kwargs.keys())
    self = self.browse(cr, uid, context=ctx)
    result = self.search(domain, offset=offset, limit=limit,
                         order=order)
    if ensure_list:
        return list(result)  # note: is result is empty returns the empty list
    else:
        return result


def field_value(self, cr, uid, ids, field_name, context=None):
    '''Read a field value for a set of objects.

    This method is very protective, if any ``False`` is passed as `ids`,
    ``False`` is returned without raising errors; also related "2one" field
    values are returned only as id integers, not tuples (id, 'name').

    :param self: model to operate in

    :param cr: database cursor

    :param uid: current user id

    :param ids: id or list of the ids of the records to read

    :param field_name: field name to return

    :param context: optional context dictionary - it may contains keys for
                    specifying certain options like ``context_lang``,
                    ``context_tz`` to alter the results of the call.

                    See method :func:`read` for more details.

    :return: a value if only one id is specified or a dictionary if more than
             one. ``False`` is returned when no record is associated with the
             ids.

    :rtype: same type of field or ``{id: value, ...}``

    :raise AccessError:

       * if user has no read rights on the requested object

       * if user tries to bypass access rules for read on the requested object

    '''
    if ids:
        fix = lambda v: v[0] if type(v) is tuple else v
        data = self.read(cr, uid, ids, [field_name], context=context)
        dt = type(data)
        if dt is dict:
            return fix(data.get(field_name, False))
        elif dt is list:
            count = len(data)
            if count == 0:
                return False
            elif count == 1:
                return fix(data[0][field_name])
            else:
                return {item['id']: fix(item[field_name]) for item in data}
        else:
            return data
    else:
        return False


def search_value(self, cr, uid, args, field_name, context=None):
    '''Similar to `find_value`:func: but searching.

    Instead of an `ids` it expects a search domain.  Matching ids will be then
    passed to `find_value`:func:.

    '''
    ids = self.search(cr, uid, args, context=context)
    return field_value(self, cr, uid, ids, field_name, context=context)


def obj_ref(self, cr, uid, xml_id):
    '''Returns (model_name, res_id) corresponding to a given `xml_id`

    If `xml_id` contains a dot '.', absolute ID (``<module>.<external-id>``)
    is assumed.

    If no module part is given, the resulting object reference is looked up in
    all alternatives with the same local ``xml-id``, first using
    ``self._module`` (for field ``module``) and after ``self._name`` (for
    field ``model``).

    If no alternative exactly match, a list of tuples will be returned, each
    one with ``(model, res_id)``.

    If no record is found ``False`` is returned.

    :param self: model to operate in

    :param cr: database cursor

    :param uid: current user id

    :param xml_id: external id to look for

    '''
    imd = self.pool.get('ir.model.data')
    fields = ['name', 'module', 'model', 'res_id']
    fname, fmod, fmodel, fid = fields
    if '.' in xml_id:
        module, ext_id = xml_id.split('.', 1)
        query = [(fmod, '=', module), (fname, '=', ext_id)]
        data = search_read(imd, cr, uid, query, [fmodel, fid])
    else:
        query = [(fname, '=', xml_id)]
        data = search_read(imd, cr, uid, domain=query, fields=fields)
    if data:
        dt = type(data)
        if issubclass(dt, dict):
            return data[fmodel], data[fid]
        elif issubclass(dt, list):
            count = len(data)
            if count == 1:
                res = data[0]
                return res[fmodel], res[fid]
            else:
                test = lambda d: d['module'] == self._module
                res = next((d for d in data if test(d)), None)
                if res:
                    return res[fmodel], res[fid]
                else:
                    test = lambda d: d[fmodel] == self._name
                    res = [(d[fmodel], d[fid]) for d in data if test(d)]
                    if not res:
                        res = [(d[fmodel], d[fid]) for d in data]
                    return res[0] if len(res) == 1 else res
    else:
        return bool(data)


def touch_fields(self, cr, uid, ids, only=None, context=None):
    '''Touches functional fields of the given `ids`.

    For each stored functional field the value is recalculated and stored.

    If `only` is set, it should be a list of fields to touch.  A single string
    is also valid.  Invalid (non-functional, or without a non-bool``store``)
    fields are silently ignored.

    If `ids` is empty, all items are touched.

    .. warning:: Don't rely in ids return from ``search()``.

       This function deals directly with the DB bypassing the ORM's
       ``write()`` method.  So you must provide ids as they are found in the
       DB.

       For instance, the ``crm_meeting`` returns many `str` IDs for each
       instance of a recurrent event though in the DB there's a single row.
       (Remember ``crm_meeting`` is actually the place for every event.)

    '''
    from xoutil.eight import iteritems, string_types
    from xoutil.names import nameof
    from xoutil.future.types import is_collection
    # FIX: This 'function' doesn't exist any more.
    from openerp.osv.fields import function
    if not ids:
        # Don't use self.search() here!  search() might return invalid ids
        # (crm.meeting does), and since _store_set_values bypasses write() and
        # executes SQL directly we must provide DB-level ids and not
        # model-levels ones.
        query = 'SELECT "id" FROM "%s"' % self._table
        cr.execute(query)
        ids = [row[0] for row in cr.fetchall()]
    if isinstance(only, string_types):
        only = (only, )
    if only is not None and not is_collection(only):
        msg = "Invalid type '%s' for argument 'only'"
        raise TypeError(msg % nameof(only, inner=True, typed=True))
    # TODO: use both _columns and _inherit_fields
    fields = [name for name, field in iteritems(self._columns)
              if not only or name in only
              if isinstance(field, function) and field.store]
    if fields:
        return self._store_set_values(
            cr, uid, ids, fields, dict(context or {})
        )


from xoeuf.osv.writers import ORMWriter as get_writer  # noqa
orm_writer = get_writer


from xoeuf.osv.writers import ORMCreator as get_creator  # noqa
orm_creator = get_creator


def cascade_search(self, cr, uid, *queries, **options):
    '''Perform a cascading search.

    Each `query` is tried in order until one returns non-empty.  The only
    keyword argument allowed is `context`.

    '''
    from collections import deque
    context = options.pop('context', {})
    if options:
        raise TypeError('Invalid keyword arguments %s' % options.popitem()[0])
    result = []
    queries = deque(queries)
    while not result and queries:
        query = queries.popleft()
        result = self.search(cr, uid, query, context=context)
    return result


def get_treeview_action(self, *args, **kwargs):
    '''Return the tree view action for `self`.

    There are two possible signatures:

    - ``get_treeview_action(recordset, **options)``

    - ``get_treeview_action(model, cr, uid, ids, context=context, **options)``

       .. warning:: ``context`` must be a keyword argument.

    If the recordset is a singleton, return the view type 'form' so that it's
    easier to see the full object.

    :rtype: An action `dict` you can return to the web client.

    `options` is used to override **almost any** key in the returned dict.  An
    interesting option is `target`.

    '''
    if args:
        (cr, uid, ids), args = args[:3], args[3:]
        if args:
            raise TypeError('Too many positional arguments')
        context = kwargs.pop('context', None)
    else:
        cr, uid, context = self.env.args
        ids = self.ids
    options = kwargs
    if not ids or len(ids) > 1:
        vtype = 'list'
        active_id = None
    else:
        vtype = 'form'
        active_id = ids[0]
    result = {
        'type': 'ir.actions.act_window',
        'res_model': self._name,
        'src_model': False,
        'view_id': False,
        'view_type': vtype,
        'view_mode': 'list,form',
        'views': [(False, 'list'), (False, 'form')],
        'target': 'current',
        'context': context,
    }
    if ids:
        result['domain'] = [('id', 'in', tuple(ids))]
    if active_id:
        result['active_id'] = active_id
    result.update(options)
    return result


del deprecated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from collections import MutableMapping
from xoutil.symbols import Unset
from xoutil.collections import SmartDictMixin


# TODO: Use `OpenDictMixin`
class ModelsManager(MutableMapping, SmartDictMixin):
    '''Xœuf models manager for a particular `registry` database.

    The mapping is essentially a mapping between model names and model
    instances.

    There is only one manager instance per database.

    Instances are:

     * A dictionary like object containing all database modules.

     * An open dictionary allowing access to keys as attributes.

    '''
    from xoutil.future.collections \
        import opendict as __search_result_type__  # noqa
    # see the the SmartDictMixin.search method

    def __new__(cls, registry):
        '''Create, or return if already exists, a instance of a models manager.
        '''
        self = super(ModelsManager, cls).__new__(cls)
        self._registry = registry
        self._mapping = None    # for "__invert__" operator cache
        return self

    def __str__(self):
        args = (type(self).__name__, self._registry.db_name, len(self))
        return str('<Database %s for "%s" with %s models>' % args)

    def __repr__(self):
        name = type(self).__name__
        db = self._registry.db_name
        return str('<Database %s for "%s">' % (name, db))

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return (isinstance(other, ModelsManager) and
                self._registry == other._registry)

    @property
    def wrapped(self):
        return self._registry.wrapped.models

    def __dir__(self):
        attrs = set(getattr(self, '__dict__', {}))
        all_slots = (getattr(cls, '__slots__', ()) for cls in type(self).mro())
        slot_attrs = {name for slots in all_slots for name in slots}
        return list(set(~self) | attrs | slot_attrs)

    def __invert__(self):
        '''Return a inverted mapping between model names (keys are identifiers
        for attributes, values are normal dotted model names).

        To obtain this mapping you can use as the unary operator "~".

        '''
        if not self._mapping or len(self) != len(self._mapping):
            self._mapping = {str(key.replace('.', '_')): key for key in self}
        return self._mapping

    def __getattr__(self, name):
        model_name = self._is_model_name(name)
        if model_name:
            return self[model_name]
        else:
            try:
                return super(ModelsManager, self).__getattr__(name)
            except Exception:  # FIXME: Should we deal with this?
                msg = "'%s' object has no attribute '%s'"
                raise AttributeError(msg % (type(self).__name__, name))

    def __setattr__(self, name, value):
        model_name = self._is_model_name(name)
        if model_name:
            self[model_name] = value
        else:
            super(ModelsManager, self).__setattr__(name, value)

    def __delattr__(self, name):
        model_name = self._is_model_name(name)
        if model_name:
            del self[name]
        else:
            super(ModelsManager, self).__delattr__(name)

    def __len__(self):
        return len(self.wrapped)

    def __iter__(self):
        return iter(self.wrapped)

    def __contains__(self, model_name):
        return model_name in self.wrapped

    def __getitem__(self, model_name):
        res = self.get(model_name, Unset)
        if res is not Unset:
            return res
        else:
            raise KeyError(model_name)

    def __setitem__(self, model_name, model):
        from xoutil.validators.identifiers import is_valid_full_identifier
        if is_valid_full_identifier(model_name):
            _valid_model_base(model)
            self.wrapped[model_name] = model
        else:
            msg = 'Inappropriate model name "%s"!'
            raise ValueError(msg % model_name)

    def __delitem__(self, model_name):
        del self.wrapped[model_name]

    def get(self, model_name, *args):
        '''Return a model for a given name or None if it doesn't exist.

        :params args: allows 0 or one value for default definition.'''
        return self._get_pop(self.wrapped.get, model_name, *args)

    def pop(self, model_name, *args):
        '''Remove specified model and return the corresponding value.

        :params args: allows 0 or one value for default definition.

        If model is not found, default value is returned if given,
        otherwise KeyError is raised.

        '''
        return self._get_pop(self.wrapped.pop, model_name, *args)

    def popitem(self):
        '''Remove and return some (model_name, model) pair as a 2-tuple;
        but raise KeyError if empty.

        '''
        return self.wrapped.popitem()

    def clear(self):
        '''Remove all models.'''
        self.wrapped.clear()

    def update(self, *args, **kwargs):
        '''A really smart update implementation.

          * Convert model name from identifiers to normal names.

          * Check that all values are valid OpenERP models.

        '''
        from collections import Mapping
        from itertools import chain
        from xoutil.validators.identifiers import is_valid_identifier
        args = [
            ((key, m[key]) for key in m)
            if isinstance(m, Mapping) else m
            for m in args
        ]
        args.append(((key, kwargs[key]) for key in kwargs))
        mapping = ~self
        for name, model in chain(args):
            if is_valid_identifier(name):
                name = mapping[name]
            self[name] = model

    def setdefault(self, model_name, default=None):
        '''return registry.get(model_name, default),
        also set registry[model_name]=default if model_name not in registry
        and default is not None

        '''
        from xoutil.validators.identifiers import is_valid_identifier
        if is_valid_identifier(model_name):
            mapping = ~self
            model_name = mapping[model_name]
        res = self.get(model_name, Unset)
        if res is Unset:
            res = default
            if default is not None:
                self[model_name] = default
        return res

    def _is_model_name(self, name):
        '''Avoid unnecessary mapping calling.

        Return a correct model name or None.

        '''
        if name not in {'db_name', 'uid'} and not name.startswith('_'):
            mapping = ~self
            return mapping.get(name)
        else:
            return None

    @staticmethod
    def _is_model_value(value):
        if value in (None, False, Unset):
            return value
        else:
            _valid_model_base(value)
            return value

    @staticmethod
    def _get_pop(method, model_name, *args):
        '''Local method used in `get` and `pop`.

        :params args: allows 0 or one value for default definition.

        '''
        count = len(args)
        if count == 0:
            res = method(model_name)
        elif count == 1:
            res = method(model_name, args[0])
        else:
            msg = '%s method expected at most 2 arguments, got "%s"!'
            raise TypeError(msg % (method.__name__, count + 1))
        return ModelsManager._is_model_value(res)


def _valid_model_base(model):
    '''Check if a model has a right base class.'''
    from xoeuf.odoo.models import BaseModel
    model_types = (BaseModel, )
    try:
        from xoeuf.odoo.models import MetaModel
        model_types += (MetaModel, )
    except ImportError:
        pass
    if not isinstance(model, model_types):
        from inspect import getmro
        from xoutil.eight import typeof
        msg = 'Inappropriate type "%s" for model value!\tMRO=%s'
        t = typeof(model)
        raise TypeError(msg % (t.__name__, getmro(t)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf Models registry for OpenERP databases.

Through this registry you can obtain a database from a customized
configuration.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from types import ModuleType
from collections import MutableMapping
from xoutil.symbols import Unset
from xoutil.context import Context
from xoutil.decorator import aliases, memoized_property

from .transaction import TransactionManager

from xoeuf.odoo import SUPERUSER_ID
# TODO: Homogenize 'manager_get' and 'manager_lock' in a compatibility module
try:
    from odoo.modules.registry import Registry as manager
    manager_get = manager
    manager_lock = lambda: manager._lock  # noqa
except ImportError:
    from openerp.modules.registry import RegistryManager as manager
    manager_get = manager.get
    manager_lock = manager.lock


class Registry(ModuleType):
    '''Xœuf Model registry for a particular database.

    The registry is essentially a mapping between model names and model
    instances. There is one registry instance per database.

    Instances are Python modules in order to be easily used in a shell.

    '''
    instances = {}

    def __new__(cls, db_name, **kwargs):
        '''Create, or return if already exists, a instance of a database
        registry.
        '''
        import threading
        with manager_lock():
            db_name = str(db_name)
            self = cls.instances.get(db_name)    # Only one per database
            if not self:
                wrapped = manager_get(db_name)
                self = super(Registry, cls).__new__(cls, db_name)
                self.db_name = db_name
                self.wrapped = wrapped
                self._default_context = kwargs
                self.uid = SUPERUSER_ID
                self._cardinality = 1
                cls.instances[db_name] = self
            else:
                self._default_context.update(kwargs)
                self._cardinality += 1
            current_thread = threading.current_thread()
            current_thread.dbname = self.db_name
            current_thread.uid = self.uid
            return self

    def __init__(self, db_name, **kwargs):
        '''A trick to finish object creation.
        '''
        if self._cardinality == 1:
            doc = str('Model registry for database "%s".\n\n'
                      'See "%s" module documentation for more info.'
                      '' % (db_name, __name__))
            super(Registry, self).__init__(str(db_name), doc)

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return isinstance(other, Registry) and self.db_name == other.db_name

    def __repr__(self):
        return '<%s for "%s" DB>' % (type(self).__name__, self.db_name)

    def __getitem__(self, model_name):
        '''Return a model for a given name or raise KeyError if it doesn't
        exist.

        "obj[model_name]" is only a shortcut for "obj.models[model_name]".

        '''
        return self.models[model_name]

    def restart(self):
        '''Restart the database connection.

        Delete an existing registry and return a newly initialized registry
        with a database connection.

        '''
        cls = type(self)
        with manager_lock():
            try:
                self.wrapped = manager.new(self.db_name)
            except Exception:
                del self.wrapped
                del cls.instances[self.db_name]
                # TODO: Manage the module in "sys.modules"
                raise

    def salt_shell(self, **kwargs):
        '''Assign common shell global variables.

        Also fix documentations and execute special tools
        ``fix_documentations`` and ``integrate_extensions``.

        '''
        from sys import _getframe
        from xoeuf.osv.improve import fix_documentations, integrate_extensions
        ENV_NAME = str('env')
        clear_names = (ENV_NAME, )
        names = ()
        f = _getframe(1)
        vars = f.f_locals
        for name in clear_names:
            if name in vars:
                var = vars[name]
                try:
                    var.clear()
                except Exception:
                    pass
        for name in names:
            var = getattr(self, name, Unset)
            if var is not Unset:
                vars[name] = var
        vars[ENV_NAME] = env = self.env
        if not kwargs:
            kwargs['_'] = 'res.users'
        if kwargs:
            models = self.models
            for kwname in kwargs:
                model_name = kwargs[kwname]
                model = models.get(model_name, Unset)
                if not model:
                    model = getattr(models, model_name, Unset)
                if model:
                    var_name = 'self' if kwname == '_' else kwname
                    # Instead of getting the model from 'models' do it from
                    # the environment.
                    vars[var_name] = env[model_name]
        fix_documentations(self)
        integrate_extensions()

    @staticmethod
    def get_all_db_names():
        '''Return all database names presents in the connected host.'''
        from xoeuf.odoo.service.db import exp_list
        return exp_list()

    @aliases('db')
    @property
    def connection(self):
        res = getattr(self.wrapped, 'db', None)
        if not res:
            res = getattr(self.wrapped, '_db')
        return res

    def __call__(self, **kwargs):
        '''Create a execution context.

        '''
        return TransactionManager(self, **kwargs)

    @property
    def cursor(self):
        '''Return a not managed cursor, callers are responsible of
        transactions and of closing it.

        '''
        return self.wrapped.cursor()

    cr = cursor    # An alias

    @memoized_property
    def models(self):
        from .models import ModelsManager
        return ModelsManager(self)

    @property
    def env(self):
        from xoeuf.odoo.api import Environment
        return Environment(self.cr, self.uid, self.context)

    @property
    def context_name(self):
        return str('%s-%s' % (__name__.partition('.')[0], self.db_name))

    @property
    def current_manager(self):
        '''Obtain the current cursor execution context if one exits, else
        return None.
        '''
        ctx = TransactionManager[self.context_name]
        return ctx or None

    @property
    def context(self):
        '''Build a context to be used as argument in Open Object methods using
        database registry defaults and current cursor context arguments; also
        checks the resulting context for common values (lang, tz).

        '''
        from xoutil.future.collections import opendict
        res = opendict(**self._default_context)
        mngr = self.current_manager
        if mngr:
            res.update(**mngr)
        return self._check_context(res)

    def _check_context(self, ctx):
        from xoeuf.odoo import api
        # TODO: catch these values
        if 'lang' not in ctx:
            import os
            DEFAULT = 'en_US'
            lang = os.environ.get('LANG', DEFAULT).split('.')[0]
            # Environment language could not be installed.
            # So, check and obtain 'lang' from DB
            with self.cr as cr:
                Lang = api.Environment(cr, self.uid, {})['res.lang']
                ok = Lang.search([('code', '=', lang)])
                if not ok:
                    predicate = [('code', 'like', '%s%%' % lang[:2])]
                    langs = Lang.search(predicate, limit=1)
                    if langs:
                        lang = langs.code
                    else:
                        lang = DEFAULT
                ctx['lang'] = lang
        if 'tz' not in ctx:
            from xoutil.fs import read_file as read
            ctx['tz'] = read('/etc/timezone').strip() or 'America/Havana'
        return ctx

    def __dir__(self):
        return list({name for name in self.__dict__.keys() + dir(type(self))
                     if not name.startswith('_')})


# Discarding not neeeded globals
del Context
del MutableMapping
del ModuleType
del aliases



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf.odoo.release import version_info as ODOO_VERSION_INFO

if ODOO_VERSION_INFO[0] == 8:
    from ._v8 import Monetary
    Monetary.__module__ = __name__
else:
    from xoeuf.odoo.fields import Monetary



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''A smart pool of OpenERP database connections managed by Xœuf.

This is mainly used in a shell, although can be used in any service.

All database names are exported as attributes from this modules.

Each database can be imported as "from xoeuf.pool import <db-name>".

If "from xoeuf.pool import test" (or any other name starting with prefix
``test``) is executed and a data-base with the given name doesn't exists, then
one in registry is returned.

When all existing data-bases are exhausted for testing purposes, standard
error will be re-raised.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from os.path import splitext
from xoutil.names import strlist as slist
from xoutil.modules import modulemethod


__docstring_format__ = 'rst'
__author__ = 'med'


def db_names():
    '''Return all `OpenERP` data-base names.

    See module documentation for more info.

    '''
    # TODO: Register mechanisms for remote hosts
    from xoeuf.osv.registry import Registry
    return Registry.get_all_db_names()


@modulemethod
def __dir__(self):
    res = [self.db_names.__name__]
    res.extend(self.db_names())
    return res


class ModuleManager(object):
    '''Import Hooks for OpenERP databases.

    A database finder and loader as a module using PEP-302 (New Import
    Hooks) protocol.

    Import Hook for OpenERP databases.

    '''
    default_context = {}
    test_db_index = 0    # Index for test databases

    def __init__(self, db_name, base):
        '''Create the Module Loader.'''
        self.db_name = db_name
        self.base = base

    @classmethod
    def hook(cls, path):
        '''Return class a the module finder.'''
        if path == __path__[0]:
            return cls
        else:
            raise ImportError

    @classmethod
    def find_module(cls, name, path=None):
        '''Return a loader object if the module was found, or None if not.

        If it raises an exception, it will be propagated to the caller,
        aborting the import.

        '''
        base, _sep, db_name = name.rpartition(str('.'))
        if base == __name__:
            return cls(db_name, base)
        else:
            msg = 'We can not load DB "%s" outside "%s".'
            raise ImportError(msg % (db_name, __name__))

    def load_module(self, fullname):
        '''Returns the loaded module or raises an exception.'''
        import sys
        self._check(fullname)
        res = sys.modules.get(fullname, None)
        if res is None:
            from psycopg2 import OperationalError
            try:
                res = self._load_module()
            except OperationalError:
                res = self._load_module(test_db=self.db_name)
                if not res:
                    raise
            sys.modules[fullname] = res
        else:
            assert res is self.registry
        return res

    def get_filename(self, fullname):
        'Return the value used in attribute `__file__` for a DB module.'
        assert self._check(fullname, asserting=True)
        return str('<%s>' % fullname)

    def _load_module(self, test_db=None):
        'Internal method for loading a DB module,'
        from xoeuf.osv.registry import Registry
        if test_db is None:
            db_name = self.db_name
        else:
            dbs = db_names()
            if test_db.startswith('test') and self.test_db_index < len(dbs):
                db_name = dbs[self.test_db_index]
                self.test_db_index += 1
            else:
                db_name = None
        if db_name:
            res = Registry(db_name, **self.default_context)
            self.registry = res
            self._inner = self.registry.wrapped      # Force check database
            res.__loader__ = self
            res.__name__ = self.db_name
            res.__file__ = self.get_filename(self.fullname)
            res.__all__ = slist('db_name', 'uid', 'context', 'connection',
                                'cursor', 'models')
            return res
        else:
            return None

    @property
    def fullname(self):
        'Return full qualified name.'
        sep = str('.')
        return sep.join((self.base, self.db_name))

    def _check(self, fullname, asserting=False):
        'Check `fullname` and raise an error if not correct.'
        local = self.fullname
        if local == fullname:
            return fullname
        elif asserting:
            return False
        else:
            msg = 'Using loader for DB "%s" for manage "%s"!'
            raise ImportError(msg % (self.db_name, fullname))


__path__ = [splitext(__file__)[0]]


import sys
sys.path_hooks.append(ModuleManager.hook)
del sys, ModuleManager, splitext, modulemethod



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.context import Context
from xoutil.names import strlist as slist

# TODO: Homogenize 'manager_lock' in a compatibility module
try:
    from odoo.modules.registry import Registry as manager
    manager_lock = lambda: manager._lock  # noqa
except ImportError:
    from openerp.modules.registry import RegistryManager as manager
    manager_lock = manager.lock


# TODO: Allow to change "openerp.tools.config" per context level
#       Implement for this "push" and "pop" methods in "xoeuf.tools.config"
class TransactionManager(Context):
    '''Manages an Odoo environment.

    There's some overlap between `odoo.api.Environment`:class: and this class.

    Usage::

       >>> with TransactionManager(registry) as tm:
       ...    tm['res.users'].search([])

    Notes:

    - You may pass the keyword 'managed' to False to avoid entering a
      transaction.

      First level contexts are always managed (regardless of the value of
      `managed`).

    - The transaction manager yields itself (``tm`` in the example above).
      The underlying Odoo environment is automatically proxied.

    '''
    __slots__ = slist('_registry', '_wrapped')

    default_context = {}    # TODO: check its value

    def __new__(cls, registry, **kwargs):
        with manager_lock():
            _super = super(TransactionManager, cls)
            self = _super.__new__(cls, registry.context_name, **kwargs)
            if self.count == 0:
                self._registry = registry
                self._wrapped = None
            else:
                assert self._registry == registry and self._wrapped
            return self

    def __init__(self, registry, **kwargs):
        super(TransactionManager, self).__init__(registry.context_name,
                                                 **kwargs)

    def __getitem__(self, key):
        return self._wrapped[key]

    def __getattr__(self, attr):
        return getattr(self._wrapped, attr)

    def __enter__(self):
        ctx = super(TransactionManager, self).__enter__()
        assert ctx is self
        if not self._wrapped:
            assert self.count == 1
            self._wrapped = self._registry.env
            self._wrapped._wrapper = self
            self._wrapped.cr.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._wrapped:
            managed = self.get('managed', True)
            if managed and self.count == 1:
                self._wrapped.cr.__exit__(exc_type, exc_val, exc_tb)
                self._wrapped = None
        return super(TransactionManager, self).__exit__(exc_type, exc_val,
                                                        exc_tb)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''An alternative OpenERP mailgate that does it's stuff in the DB instead via
XMLRPC.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import time
import random

from logging import Handler
from psycopg2 import OperationalError, errorcodes

from xoutil.future.codecs import safe_encode, safe_decode

try:
    from xoeuf.odoo.jobs import Deferred
except ImportError:
    Deferred = None

# A size limit for messages sent through the message broker.
MAX_SIZE_FOR_DEFERRED = 200 * 1024

from . import Command


PG_CONCURRENCY_ERRORS_TO_RETRY = (
    errorcodes.LOCK_NOT_AVAILABLE,
    errorcodes.SERIALIZATION_FAILURE,
    errorcodes.DEADLOCK_DETECTED
)
MAX_TRIES_ON_CONCURRENCY_FAILURE = 5

try:
    # A Py3k more compatible Exception value
    # TODO: Use `xoutil.eight.exceptions.StandardError`
    #       Rename all use of `Exception`
    Exception = StandardError
except NameError:
    pass


# TODO: This has grown into a monstrous pile of code that needs
# refactorization.


CR = str('\r')
LF = str('\n')
CRLF = CR + LF


# TODO: Should this be moved elsewhere?
class SysLogHandler(Handler):
    def emit(self, report):
        # This avoids the /dev/log system issue and goes directly to Unix
        # syslog.  But then Windows is f.cked.
        import syslog
        syslog.syslog(safe_encode(self.format(report)))


del Handler


class Mailgate(Command):
    '''The xoeuf mailgate for OpenERP.

    It does not need the XMLRPC server to be running.  The only requirement is
    that you have configured properly the OpenERP:

    a) Set up DB host, user and password.

    '''

    MESSAGE_TEMPLATE = safe_encode(
        "Error while processing message."
        "from {sender}.\n\n"
        "{traceback}\n\n"
        "{details_title}:\n\n"
        "{message_details}\n"
    )

    @classmethod
    def get_arg_parser(cls):
        def path(extensions=None):
            '''A type-builder for file arguments.'''
            from xoutil.future.types import is_collection
            from os.path import abspath, isfile, splitext
            if extensions and not is_collection(extensions):
                extensions = (extensions, )
            acceptable = lambda ext: not extensions or ext in extensions

            def inner(value):
                res = abspath(value)
                name, extension = splitext(value)
                if not isfile(res) or not acceptable(extension):
                    raise TypeError('Invalid filename %r' % res)
                return res
            return inner

        res = getattr(cls, '_arg_parser', None)
        if not res:
            from argparse import ArgumentParser
            res = ArgumentParser()
            cls._arg_parser = res
            res.add_argument('--quick', dest='quick',
                             action='store_true',
                             default=False,
                             help=('Accept the message as quickly as '
                                   'possible, deferring actions if needed. '
                                   'NOTICE: This means your Odoo application '
                                   'needs to be prepared to send bounces '
                                   'if needed.'))
            res.add_argument('-c', '--config', dest='conf',
                             required=False,
                             type=path(),
                             help='A configuration file.  This could be '
                             'either a Python file, like that required by '
                             'Gunicorn deployments, or a INI-like '
                             'like the standard ".openerp-serverrc".')
            res.add_argument('-d', '--database', dest='database',
                             required=True)
            res.add_argument('-m', '--model', dest='default_model',
                             default=str('crm.lead'),
                             type=str,
                             help='The fallback model to use if the message '
                             'it not a reply or is not addressed to an '
                             'email alias.  Defaults to "crm.lead", i.e. '
                             'creating a CRM lead.')
            res.add_argument('--strip-attachment', dest='strip_attachments',
                             default=False,
                             action='store_true',
                             help='Set to strip the attachments from the '
                             'messages.')
            res.add_argument('--save-original', dest='save_original',
                             default=False,
                             action='store_true',
                             help='Set to also save the message in '
                             'its original format.')
            res.add_argument('--slowness', dest='slowness',
                             default=0,
                             type=float,
                             help='How much time in seconds to wait '
                             'before standard input needs to be ready. '
                             'Defaults to 0 (i.e not to wait).')
            res.add_argument('--allow-empty', dest='allow_empty',
                             default=False,
                             action='store_true',
                             help='Whether to accept an empty message '
                             'without error.')
            res.add_argument('--input-file', '-f', dest='input',
                             type=path(),
                             help='Read the message from the file INPUT '
                             'instead of the stdin.')
            res.add_argument('--defer', default=False, action='store_true',
                             help='Treat errors as transient.')
            res.add_argument('--queue-id', dest='queue_id', default='',
                             help=('The queue ID the MTA queue.  If provided '
                                   'the header X-Queue-ID will be injected '
                                   'to the message'))
            # Deprecated, but simply ignored
            loggroup = res.add_argument_group('Logging')
            loggroup.add_argument('--log-level',
                                  choices=('debug', 'warning',
                                           'info', 'error'),
                                  default='warning',
                                  help='How much to log')
            loggroup.add_argument('--log-host', default=None,
                                  help='The SMTP host for reporting errors.')
            loggroup.add_argument('--log-to', default=None,
                                  nargs='+',
                                  help='The address to receive error logs.')
            loggroup.add_argument('--log-from', default=None,
                                  help='The issuer of error reports.')
        return res

    @classmethod
    def database_factory(cls, database):
        # TODO: Homogenize 'get' in a compatibility module.
        try:
            from odoo.modules.registry import Registry
            get = Registry
        except ImportError:
            from openerp.modules.registry import RegistryManager
            get = RegistryManager.get
        return get(database)

    @staticmethod
    def get_raw_message(timeout=0, raises=True):
        '''Return the data provide via stdin.

        This will wait until the stdin is ready until the timeout expires.  If
        the `raises` is True and the stdin returns an empty byte-stream a
        RuntimeError is raised.

        The result is always bytes.

        '''
        from xoutil.eight import binary_type as bytes
        import select
        import sys
        import logging
        logger = logging.getLogger(__name__)
        ready, _, _ = select.select([sys.stdin], [], [], timeout)
        if ready:
            stdin = ready[0]
            # XXX: To read bytes in both Py3k and Py 2.
            buffer = getattr(stdin, 'buffer', stdin)
            result = buffer.read()
            return result
        elif raises:
            raise RuntimeError('No message via stdin')
        else:
            logger.warn('No message provided, but allowing.')
            return bytes('')

    @classmethod
    def send_error_notification(cls, message):
        '''Report an error dealing with `message`.'''
        import logging
        import email
        from email.message import Message
        try:
            logger = logging.getLogger(__name__)
            if not isinstance(message, Message):
                msg = email.message_from_string(safe_encode(message))
            else:
                msg = message
                message = safe_encode(msg.as_string())
            # The follow locals vars are mean to be retrieved from tracebacks.
            msgid = safe_encode(msg.get('Message-Id', '<NO ID>'))  # noqa
            mail_from = safe_encode(msg.get('From', '<nobody>'))  # noqa
            sender = safe_encode(msg.get('Sender', '<nobody>'))  # noqa
            logger.critical("Error while processing incoming message.",
                            exc_info=1)
        except Exception:
            # Avoid errors... This should be logged to the syslog instead and
            # raven prints the connection error.
            pass

    def send_immediate(self, options, message):
        from xoeuf.odoo import SUPERUSER_ID, api
        default_model = options.default_model
        db = self.database_factory(options.database)
        with db.cursor() as cr:
            env = api.Environment(cr, SUPERUSER_ID, {})
            obj = env['mail.thread']
            obj.message_process(default_model,
                                message,
                                save_original=options.save_original,
                                strip_attachments=options.strip_attachments)

    def send_deferred(self, options, message):
        from xoeuf.odoo import SUPERUSER_ID
        default_model = options.default_model
        Deferred(
            str('mail.thread'),
            options.database, SUPERUSER_ID, 'message_process',
            default_model, message, save_original=options.save_original,
            strip_attachments=options.strip_attachments
        )

    def run(self, args=None):
        from xoeuf.sentrylog import patch_logging
        patch_logging(override=True, force=True)
        parser = self.get_arg_parser()
        options = parser.parse_args(args)
        conffile = options.conf
        if conffile:
            self.read_conffile(conffile)
        message = None
        try:
            if not options.input:
                message = self.get_raw_message(timeout=options.slowness,
                                               raises=not options.allow_empty)
            else:
                with open(options.input, 'rb') as f:
                    message = f.read()
            # TODO: assert message is bytes
            if options.queue_id:
                message = (
                    'X-Queue-ID: %s%s' % (options.queue_id, CRLF) + message
                )
            retries = 0
            done = False
            while not done:
                try:
                    if Deferred and options.quick and random.random() < 0.1 and len(message) < MAX_SIZE_FOR_DEFERRED:  # noqa
                        # In order to test this feature we're only deferring a
                        # 10% of messages.  Increase this number with care
                        # until you reach 100%.  Also don't send large emails
                        # through the broker, that may breaker the broker ;)
                        self.send_deferred(options, message)
                    else:
                        self.send_immediate(options, message)
                except OperationalError as error:
                    if error.pgcode not in PG_CONCURRENCY_ERRORS_TO_RETRY:
                        raise
                    if retries < MAX_TRIES_ON_CONCURRENCY_FAILURE:
                        retries += 1
                        wait_time = random.uniform(0.0, 2**retries)
                        time.sleep(wait_time)
                    else:
                        raise
                else:
                    done = True
        except Exception:
            import sys
            if options.defer:
                print(str('4.3.5 System incorrectly configured'),
                      file=sys.stderr)
            else:
                print(str('5.0.0 Permanent error. System error.'),
                      file=sys.stderr)
            # First decode the message since the raw message may contain
            # invalid UTF-8 sequences or come with a mixture of encodings
            # (each MIME part can be encoded differently) and thus the logger
            # fail to construct the log message.
            self.send_error_notification(
                safe_decode(message) if message else 'No message provided'
            )
            if sys.stdout.isatty():
                import traceback
                traceback.print_exc()
            sys.exit(1)

    def read_conffile(self, filename):
        import os
        ext = os.path.splitext(filename)[-1]
        if ext == '.py':
            self.load_config_from_script(filename)
        else:
            self.load_config_from_inifile(filename)

    @staticmethod
    def load_config_from_script(filename):
        from xoutil.eight import exec_
        cfg = {
            "__builtins__": __builtins__,
            "__name__": "__config__",
            "__file__": filename,
            "__doc__": None,
            "__package__": None
        }
        try:
            with open(filename, 'rb') as fh:
                return exec_(compile(fh.read(), filename, 'exec'), cfg, cfg)
        except Exception:
            import traceback
            import sys
            print("Failed to read config file: %s" % filename)
            traceback.print_exc()
            sys.exit(1)

    @staticmethod
    def load_config_from_inifile(filename):
        from xoeuf.odoo.tools import config
        config.rcfile = filename
        config.load()


def main():
    '''The OpenERP mailgate.'''
    from xoutil.cli.app import main
    from xoutil.cli import command_name
    main(default=command_name(Mailgate))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from operator import attrgetter
from xoeuf.odoo.fields import Field

import xoeuf.odoo.osv.fields as fields7

if not hasattr(fields7, 'monetary'):
    import __builtin__

    class monetary(fields7.float):
        _type = 'monetary'
        _symbol_set = ('%s', lambda x: __builtin__.float(x or 0.0))
        _symbol_get = lambda self, x: x or 0.0  # noqa: E731

        def to_field_args(self):
            raise NotImplementedError(
                "fields.monetary is only supported in the new API, "
                "but you can use widget='monetary' in client-side views"
            )

    fields7.monetary = monetary
    del monetary

del fields7


class Monetary(Field):
    """The decimal precision and currency symbol are taken from the attribute

    :param currency_field: name of the field holding the currency this monetary
                           field is expressed in (default: `currency_id`)
    """
    type = 'monetary'
    _slots = {
        'currency_field': None,
        'group_operator': None,         # operator for aggregating values
    }

    def __init__(self, string=None, currency_field=None, **kwargs):
        super(Monetary, self).__init__(
            string=string,
            currency_field=currency_field,
            **kwargs
        )

    _related_currency_field = property(attrgetter('currency_field'))
    _related_group_operator = property(attrgetter('group_operator'))

    _description_currency_field = property(attrgetter('currency_field'))
    _description_group_operator = property(attrgetter('group_operator'))

    _column_currency_field = property(attrgetter('currency_field'))
    _column_group_operator = property(attrgetter('group_operator'))

    def _setup_regular(self, env):
        super(Monetary, self)._setup_regular(env)
        model = env[self.model_name]
        if not self.currency_field:
            # pick a default, trying in order: 'currency_id', 'x_currency_id'
            if 'currency_id' in model._fields:
                self.currency_field = 'currency_id'
            elif 'x_currency_id' in model._fields:
                self.currency_field = 'x_currency_id'
        assert self.currency_field in model._fields, \
            "Field %s with unknown currency_field %r" % (self, self.currency_field)   # noqa

    def convert_to_cache(self, value, record, validate=True):
        currency = record[self.currency_field]
        if currency:
            return currency.round(float(value or 0.0))
        return float(value or 0.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoeuf import models, fields
from xoeuf.odoo import signals


class SignalingModel(models.Model):
    _name = 'test_signals.signaling_model'

    name = fields.Char()


@signals.receiver(signals.post_save, sender='test_signals.signaling_model')
def do_nothing(sender, signal, **kwargs):
    pass


@signals.receiver(signals.post_save, sender='test_signals.signaling_model')
def do_nothing2(sender, signal, **kwargs):
    pass


@signals.receiver(signals.post_save)
def do_nothing3(sender, signal, **kwargs):
    pass


@signals.receiver(signals.pre_save, sender='test_signals.signaling_model')
def do_nothing_again(sender, signal, **kwargs):
    pass


@signals.wrapper(signals.write_wrapper, sender='test_signals.signaling_model')
def wrap_nothing(sender, wrapping, *args, **kwargs):
    yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf extensions for Open Object (OpenERP) models.

This module define functions to improve `OpenERP` object services (OSV) with
some extensions related to model programming or shell (Command Line Interface)
use:

- :func:`integrate_extensions` -  integrate all methods defined as
  functions in module ``xoeuf.osv.model_extensions`` to `ModelBase`.

- :func:`fix_documentations` - Fixes all models documentation from a given
  Xœuf registry (`OpenERP` data-base).

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import)

from xoutil.names import strlist as strs


__all__ = strs('integrate_extensions', 'fix_documentations')

del strs


def integrate_extensions():
    '''Integrate all functions defined in ``xoeuf.osv.model_extensions`` as new
    `BaseModel` methods and operators.

    It can be used in Python modules like::

      from xoeuf.osv.improve import integrate_extensions as _ie
      _ie(); del _ie

    '''
    from types import FunctionType
    from xoeuf.odoo.models import BaseModel
    from xoeuf.models import extensions as model_extensions
    from xoeuf.api import guess as adapt

    def fixname(name):
        'Convert operator names, just in case'
        op_prefix = 'operator__'
        c = len(op_prefix)
        return '__%s__' % name[c:] if name.startswith(op_prefix) else name

    if not model_extensions.INTEGRATED:
        model_extensions.INTEGRATED = True
        for name in dir(model_extensions):
            value = getattr(model_extensions, name)
            if type(value) is FunctionType:
                setattr(BaseModel, fixname(name), adapt(value))


def fix_documentations(db):
    '''Fixes all models documentation from a given data-base.

    This function may be useful for shells or Python Command Line Interfaces
    (CLI).

    '''
    from xoutil.objects import (fix_class_documentation,
                                fix_method_documentation)
    if not getattr(db, '__documentation_fixed', False):
        setattr(db, '__documentation_fixed', True)
        ignore = ('object', 'AbstractModel', 'BaseModel', 'Model',
                  'TransientModel')
        models = db.models
        for model in models.values():
            cls = model.__class__
            fix_class_documentation(cls, ignore=ignore, deep=10)
            for attr_name in dir(cls):
                fix_method_documentation(cls, attr_name, deep=3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


import hypothesis
from hypothesis import strategies as s
from math import isinf
from xoeuf.odoo.tests.common import TransactionCase

finite_floats = s.floats(allow_nan=False, allow_infinity=False)


class TestMonetary(TransactionCase):
    @hypothesis.given(finite_floats)
    def test_monetary_value(self, value):
        Line = self.env['test.monetary.line']
        obj = Line.create({'value': value})

        # Big-big values causes obj.value to be inf, skip those
        hypothesis.assume(not isinf(obj.value))

        # For big numbers like ``value=3.6893488147419103e+19``, we get
        # approximate ``obj.value=3.68934881474191e+19`` which, when
        # substracted is quite large (4096); this represents a loss of
        # precision representing these values and it may be introduced either
        # by orm, psycopg2 or postgresql (check who).  According to docs both
        # Python and PG use 64 bits for floats (53 bits of precision), I don't
        # know why I can represent that value in Python but we something
        # different
        #
        # Since assertAlmostEqual does ``round(a - b, e) == 0`` this
        # difference is large enough to fail.  To overcome that I do
        # ``round(a/b - 1, 7) == 0``.
        if value:
            self.assertAlmostEqual(
                obj.value / value, 1,
                places=7,
                msg='%s !~ %s from %s' % (obj.value, value, obj.value / value)
            )
        else:
            self.assertAlmostEqual(obj.value, value, places=7)

        cr = self.env.cr
        cr.execute(
            'SELECT value FROM test_monetary_line WHERE id=%s',
            (obj.id, )
        )
        fetched = cr.fetchone()[0]
        if value:
            self.assertAlmostEqual(
                fetched / value, 1,
                places=7,
                msg='%s !~ %s from %s' % (fetched, value, fetched / value)
            )
        else:
            self.assertAlmostEqual(fetched, value, places=7)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf Configuration Service.

Define the class :class:`MetaOptions` which is a singleton that wraps
``openerp.tools.config``, it's instanced automatically at the end of this
module.

If Xœuf Services are called from an application with arguments, they are parsed
and loaded the first time this module is used.

If a command is invoked, it's looked in all OpenERP configured modules (addons)
and its arguments are parsed and it's executed.

You can update or load all options by calling in any time either the method
:meth:`!options.load` or the method :meth:`!options.update`.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from collections import MutableMapping
from xoutil.eight.meta import metaclass


DEFAULT_COMMAND = str('server')
_SECTION_SEP = str('.')


class MetaOptions(type):
    __singleton__ = None

    def __new__(cls, name, bases, attrs):
        if cls.__singleton__ is None:
            from xoeuf.odoo.tools import config
            attrs['__new__'] = None     # can't be instantiated!
            attrs['wrapped'] = config
            self = super(MetaOptions, cls).__new__(cls, name, bases, attrs)
            cls.__singleton__ = self
            return self
        else:
            msg = ('Only one configuration instance allowed:\n'
                   '\tdefining "%s", already defined "%s"!')
            TypeError(msg % (name, cls.__singleton__.__name__))

    def __dir__(self):
        return (list(self.wrapped.options) + list(self.__dict__))

    def __hash__(self):
        from xoutil.names import nameof
        return hash(nameof(self, inner=True, full=True))

    def __eq__(self, other):
        return other is self

    def __ne__(self, other):
        return not (self == other)

    def __len__(self):
        res = len(self.wrapped.options)
        for section in self.wrapped.misc:
            res += len(self.wrapped.misc[section])
        return res

    def __contains__(self, option):
        if _SECTION_SEP in option:
            section, option = option.split(_SECTION_SEP)
            return option in self.wrapped.misc.get(section, {})
        else:
            return option in self.wrapped.options

    def __iter__(self):
        for option in self.wrapped.options:
            yield option
        for section in self.wrapped.misc:
            for option in self.wrapped.misc[section]:
                yield _SECTION_SEP.join((section, option))

    def __getitem__(self, option):
        if _SECTION_SEP in option:
            section, option = option.split(_SECTION_SEP)
            return self.wrapped.misc[section][option]
        else:
            return self.wrapped[option]

    def __setitem__(self, option, value):
        from xoutil.eight import string_types
        if isinstance(option, string_types):
            option = str(option)
            if value in ('True', 'true'):
                value = True
            elif value in ('False', 'false'):
                value = False
            elif value == 'None':
                value = None
            if _SECTION_SEP in option:
                section, option = option.split(_SECTION_SEP)
                self.wrapped.misc[section][option] = value
            else:
                self.wrapped[option] = value
        else:
            from xoutil.eight import typeof
            msg = 'option name must be str, "%s" of type "%s" is provided!'
            raise TypeError(msg % (option, typeof(option).__name__))

    def __delitem__(self, option):
        if _SECTION_SEP in option:
            section, option = option.split(_SECTION_SEP)
            del self.wrapped.misc[section][option]
        elif option not in self.wrapped.casts:
            del self.wrapped.options[option]
        else:
            raise KeyError('read only option "%s"!' % option)

    def __getattr__(self, name):
        from xoutil.symbols import Unset
        res = self.wrapped.options.get(name, Unset)
        if res is not Unset:
            return res
        else:
            msg = "'%s' has no attribute '%s'"
            raise AttributeError(msg % (self.__name__, name))

    def __setattr__(self, name, value):
        if name not in self.__dict__ and name in self.wrapped.options:
            self[name] = value
        else:
            super(MetaOptions, self).__setattr__(name, value)

    def get(self, option, default=None):
        '''returns options[option] if option in options, else default

            default is None if not given.

        '''
        if _SECTION_SEP in option:
            from xoutil.symbols import Unset
            section, option = option.split(_SECTION_SEP)
            misc = self.wrapped.misc.get(section, Unset)
            return default if misc is Unset else misc.get(option, default)
        else:
            return self.wrapped.options.get(option, default)

    def keys(self):
        '''return a set-like object providing a view on options' keys'''
        from collections import KeysView
        return KeysView(self)

    def items(self):
        '''return a set-like object providing a view on options' items'''
        from collections import ItemsView
        return ItemsView(self)

    def values(self):
        '''return an object providing a view on options' values'''
        from collections import ValuesView
        return ValuesView(self)

    def pop(self, option, *args):
        '''options.pop(option[,default]) -> remove specified option name and
        return the corresponding option value.

        If option is not found, default is returned if given, otherwise
        KeyError is raised.
        '''
        count = len(args)
        if count <= 1:
            from xoutil.symbols import Unset
            default = Unset if count == 0 else args[0]
            if _SECTION_SEP in option:
                section, option = option.split(_SECTION_SEP)
                misc = self.wrapped.misc.get(section, Unset)
                if misc is not Unset:
                    if default is Unset:
                        return misc.pop(option)
                    else:
                        return misc.pop(option, default)
                else:
                    raise KeyError(option)
            else:
                if default is Unset:
                    return self.wrapped.options.pop(option)
                else:
                    return self.wrapped.option.pop(option, default)
        else:
            msg = 'pop expected at most 2 arguments, got %s'
            raise TypeError(msg % count + 1)

    def update(self, *args, **kwargs):
        '''Update options from dict/iterable or keyword arguments

        If a positional argument (other) is present:

          - If has a "keys" method, does::
            for option in other:
                options[option] = other[option]

          - If not, does::
            for option, value in other:
                options[option] = value

        In either case, this is followed by::
            for option in kwargs:
                options[option] = kwargs[option]

        '''
        count = len(args)
        if count <= 1:
            if args:
                other = args[0]
                if hasattr(other, 'keys'):
                    for option in other:
                        self[option] = other[option]
                else:
                    for option, value in other:
                        self[option] = value
            for option in kwargs:
                self[option] = kwargs[option]
        else:
            msg = 'update expected at most 1 positional argument, got %s'
            raise TypeError(msg % count)

    def setdefault(self, option, *args):
        '''return options.get(`option`, `default`)

        also set options[`option`] = `default` if `option` not in options

        if `default` is not given, will try to find the default value
        configured, is not an standard option, then will use None.
        '''
        count = len(args)
        if count <= 1:
            from xoutil.symbols import Unset
            default = Unset if count == 0 else args[0]
            if _SECTION_SEP in option:
                default = None if count == 0 else args[0]
                section, option = option.split(_SECTION_SEP)
                return self.wrapped.misc[section].setdefault(option, default)
            else:
                if count == 0:
                    opt = self.wrapped.casts.get(option)
                    default = opt.my_default if opt else None
                else:
                    default = args[0]
                return self.wrapped.options.setdefault(option, default)
        else:
            msg = 'setdefault expected at most 2 arguments, got %s'
            raise TypeError(msg % count + 1)


MutableMapping.register(MetaOptions)


class options(metaclass(MetaOptions)):
    '''The single instance of :class:`MetaOptions` that wraps
    ``openerp.tools.config``.

    '''


del metaclass
del MutableMapping



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Secures a DB by:

- Changing all passwords

- [TODO] Changing all accounting related data by a random mapping.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from . import Command


class Secure(Command):
    '''The secure command.

    '''

    @classmethod
    def get_arg_parser(cls):
        from xoutil.crypto import PASS_LEVEL_NAME_MAPPING
        from xoutil.crypto import DEFAULT_PASS_PHRASE_LEVEL

        def path(extensions=None):
            '''A type-builder for file arguments.'''
            from xoutil.future.types import is_collection
            from os.path import abspath, isfile, splitext
            if extensions and not is_collection(extensions):
                extensions = (extensions, )
            acceptable = lambda ext: not extensions or ext in extensions

            def inner(value):
                res = abspath(value)
                name, extension = splitext(value)
                if not isfile(res) or not acceptable(extension):
                    raise TypeError('Invalid filename %r' % res)
                return res
            return inner

        res = getattr(cls, '_arg_parser', None)
        if not res:
            from argparse import ArgumentParser
            res = ArgumentParser()
            cls._arg_parser = res
            res.add_argument('-c', '--config', dest='conf',
                             required=False,
                             default=None,
                             type=path(),
                             help='A configuration file.  This could be '
                             'either a Python file, like that required by '
                             'Gunicorn deployments, or a INI-like '
                             'like the standard ".openerp-serverrc".')
            res.add_argument('-l', '--level', dest='security_level',
                             choices=PASS_LEVEL_NAME_MAPPING.keys(),
                             default=DEFAULT_PASS_PHRASE_LEVEL,
                             help='The security level for passwords.  If '
                             '"basic" the password will the same as the '
                             'user\'s login.  "random" means a truly random '
                             'password.')
            res.add_argument('-d', '--database', dest='database',
                             type=cls.database_factory,
                             required=True)
            loggroup = res.add_argument_group('Logging')
            loggroup.add_argument('--log-level',
                                  choices=('debug', 'warning',
                                           'info', 'error'),
                                  default='warning',
                                  help='How much to log')
        return res

    @classmethod
    def database_factory(cls, database):
        from xoeuf.odoo import api, SUPERUSER_ID
        # TODO: Homogenize 'get' in a compatibility module.
        try:
            from xoeuf.odoo.modules.registry import RegistryManager
            get = RegistryManager.get
        except ImportError:
            from xoeuf.odoo.modules.registry import Registry
            get = Registry.get
        db = get(database)
        env = api.Environment(db.cursor(), SUPERUSER_ID, {})
        return env['res.users']

    def run(self, args=None):
        from xoeuf.security import reset_all_passwords
        parser = self.get_arg_parser()
        options = parser.parse_args(args)
        conffile = options.conf
        level = options.security_level
        if conffile:
            self.read_conffile(conffile)
        self = options.database
        with self.env.cr:
            reset_all_passwords(self, security_level=level)

    def read_conffile(self, filename):
        import os
        ext = os.path.splitext(filename)[-1]
        if ext == '.py':
            self.load_config_from_script(filename)
        else:
            self.load_config_from_inifile(filename)

    @staticmethod
    def load_config_from_script(filename):
        from xoutil.eight import exec_
        cfg = {
            "__builtins__": __builtins__,
            "__name__": "__config__",
            "__file__": filename,
            "__doc__": None,
            "__package__": None
        }
        try:
            with open(filename, 'rb') as fh:
                return exec_(compile(fh.read(), filename, 'exec'), cfg, cfg)
        except Exception:
            import traceback
            import sys
            print("Failed to read config file: %s" % filename)
            traceback.print_exc()
            sys.exit(1)

    @staticmethod
    def load_config_from_inifile(filename):
        from xoeuf.odoo.tools import config
        config.rcfile = filename
        config.load()
